@@ -39,6 +39,7 @@ static void gth_viewer_page_interface_init (GthViewerPageInterface *iface);
 
 struct _GthMediaViewerPagePrivate {
 	GthBrowser     *browser;
+	GSettings      *settings;
 	GthFileData    *file_data;
 	GFileInfo      *updated_info;
 	GstElement     *playbin;
@@ -47,6 +48,7 @@ struct _GthMediaViewerPagePrivate {
 	GtkWidget      *audio_area;
 	GtkWidget      *area_box;
 	GtkWidget      *area_overlay;
+	gboolean        fit_if_larger;
 	gboolean        visible;
 	gboolean        playing;
 	gboolean        paused;
@@ -54,6 +56,8 @@ struct _GthMediaViewerPagePrivate {
 	gint64          duration;
 	int             video_fps_n;
 	int             video_fps_d;
+	int             video_width;
+	int             video_height;
 	gboolean        has_video;
 	gboolean        has_audio;
 	gulong          update_progress_id;
@@ -68,6 +72,7 @@ struct _GthMediaViewerPagePrivate {
 	gboolean        cursor_visible;
 	GthScreensaver *screensaver;
 	GtkWidget      *screenshot_button;
+	GtkWidget      *fit_button;
 	gboolean        background_painted;
 };
 
@@ -84,7 +89,9 @@ static double default_rates[] = { 0.03, 0.06, 0.12, 0.25, 0.33, 0.50, 0.66, 1.0,
 
 
 static const GActionEntry actions[] = {
-	{ "video-screenshot", gth_browser_activate_video_screenshot }
+	{ "video-screenshot", gth_browser_activate_video_screenshot },
+	{ "toggle-play", gth_browser_activate_toggle_play },
+	{ "video-zoom-fit", toggle_action_activated, NULL, "true", gth_browser_activate_video_zoom_fit },
 };
 
 
@@ -165,6 +172,52 @@ video_area_unrealize_cb (GtkWidget *widget,
 }
 
 
+static void
+update_zoom_info (GthMediaViewerPage *self)
+{
+	GtkAllocation  allocation;
+	double         view_width;
+	double         view_height;
+	int            zoom;
+	char          *text;
+
+	if (! self->priv->has_video) {
+		gth_statusbar_set_secondary_text (GTH_STATUSBAR (gth_browser_get_statusbar (self->priv->browser)), "");
+		return;
+	}
+
+	gtk_widget_get_allocation (self->priv->video_area, &allocation);
+
+	view_width = allocation.width;
+	view_height = (((double) self->priv->video_height / self->priv->video_width) * view_width);
+	if (view_height > allocation.height) {
+		view_height = allocation.height;
+		view_width = (((double) self->priv->video_width / self->priv->video_height) * view_height);
+	}
+
+	if (self->priv->video_width > 0)
+		zoom = (int) round ((double) view_width / self->priv->video_width * 100);
+	else if (self->priv->video_height > 0)
+		zoom = (int) round ((double) view_height / self->priv->video_height * 100);
+	else
+		zoom = 100;
+	text = g_strdup_printf ("  %d%%  ", zoom);
+	gth_statusbar_set_secondary_text (GTH_STATUSBAR (gth_browser_get_statusbar (self->priv->browser)), text);
+
+	g_free (text);
+}
+
+
+static void
+video_area_size_allocate_cb (GtkWidget    *widget,
+			     GdkRectangle *allocation,
+			     gpointer      user_data)
+{
+	GthMediaViewerPage *self = user_data;
+	update_zoom_info (self);
+}
+
+
 static gboolean
 video_area_draw_cb (GtkWidget *widget,
 		    cairo_t   *cr,
@@ -415,39 +468,7 @@ static void
 play_button_clicked_cb (GtkButton *button,
 			gpointer   user_data)
 {
-	GthMediaViewerPage *self = user_data;
-
-	if (self->priv->playbin == NULL)
-		return;
-	if (! self->priv->playing) {
-		if (! self->priv->paused) {
-			gst_element_set_state (self->priv->playbin, GST_STATE_PAUSED);
-			gst_element_seek (self->priv->playbin,
-					  self->priv->rate,
-					  GST_FORMAT_TIME,
-					  GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_ACCURATE,
-					  GST_SEEK_TYPE_SET,
-					  0.0,
-					  GST_SEEK_TYPE_NONE,
-					  0.0);
-		}
-		else {
-			gint64 current_value;
-
-			current_value = (gint64) (gtk_adjustment_get_value (GTK_ADJUSTMENT (GET_WIDGET ("position_adjustment"))) / 100.0 * self->priv->duration);
-			gst_element_seek (self->priv->playbin,
-					  self->priv->rate,
-					  GST_FORMAT_TIME,
-					  GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_ACCURATE,
-					  GST_SEEK_TYPE_SET,
-					  current_value,
-					  GST_SEEK_TYPE_NONE,
-					  0.0);
-		}
-		gst_element_set_state (self->priv->playbin, GST_STATE_PLAYING);
-	}
-	else
-		gst_element_set_state (self->priv->playbin, GST_STATE_PAUSED);
+	gth_media_viewer_page_toggle_play (GTH_MEDIA_VIEWER_PAGE (user_data));
 }
 
 
@@ -659,6 +680,9 @@ update_stream_info (GthMediaViewerPage *self)
 		      "video-sink", &video_sink,
 		      NULL);
 
+	self->priv->has_audio = FALSE;
+	self->priv->has_video = FALSE;
+
 	if (audio_sink != NULL) {
 		audio_pad = gst_element_get_static_pad (GST_ELEMENT (audio_sink), "sink");
 		if (audio_pad != NULL) {
@@ -689,6 +713,8 @@ update_stream_info (GthMediaViewerPage *self)
 					g_file_info_set_attribute_int32 (self->priv->updated_info, "frame::width", video_width);
 					g_file_info_set_attribute_int32 (self->priv->updated_info, "frame::height", video_height);
 					self->priv->has_video = TRUE;
+					self->priv->video_width = video_width;
+					self->priv->video_height = video_height;
 				}
 
 				gst_caps_unref (caps);
@@ -697,6 +723,7 @@ update_stream_info (GthMediaViewerPage *self)
 	}
 
 	gtk_stack_set_visible_child_name (GTK_STACK (self->priv->area_box), self->priv->has_video ? "video-area" : "audio-area");
+	update_zoom_info (self);
 }
 
 
@@ -800,7 +827,6 @@ static void
 create_playbin (GthMediaViewerPage *self)
 {
 	GstElement *video_sink;
-	GSettings  *settings;
 	GstBus     *bus;
 
 	if (self->priv->playbin != NULL)
@@ -808,8 +834,7 @@ create_playbin (GthMediaViewerPage *self)
 
 	self->priv->playbin = gst_element_factory_make ("playbin", "playbin");
 
-	settings = g_settings_new (GTHUMB_GSTREAMER_TOOLS_SCHEMA);
-	if (g_settings_get_boolean (settings, PREF_GSTREAMER_USE_HARDWARE_ACCEL)) {
+	if (g_settings_get_boolean (self->priv->settings, PREF_GSTREAMER_USE_HARDWARE_ACCEL)) {
 		GstElement *video_bin;
 		GstElement *glupload;
 		GstPad     *pad;
@@ -856,26 +881,17 @@ create_playbin (GthMediaViewerPage *self)
 			  G_CALLBACK (video_area_unrealize_cb),
 			  self);
 	g_signal_connect (G_OBJECT (self->priv->video_area),
-			  "button_press_event",
-			  G_CALLBACK (video_area_button_press_cb),
-			  self);
-	g_signal_connect (G_OBJECT (self->priv->video_area),
-			  "popup-menu",
-			  G_CALLBACK (video_area_popup_menu_cb),
-			  self);
-	g_signal_connect (G_OBJECT (self->priv->video_area),
-			  "scroll_event",
-			  G_CALLBACK (video_area_scroll_event_cb),
+			  "size-allocate",
+			  G_CALLBACK (video_area_size_allocate_cb),
 			  self);
 
 	gtk_stack_add_named (GTK_STACK (self->priv->area_box), self->priv->video_area, "video-area");
 	gtk_widget_show (self->priv->video_area);
 
 	g_object_set (self->priv->playbin,
-		      "volume", (double) g_settings_get_int (settings, PREF_GSTREAMER_TOOLS_VOLUME) / 100.0,
+		      "volume", (double) g_settings_get_int (self->priv->settings, PREF_GSTREAMER_TOOLS_VOLUME) / 100.0,
 		      "force-aspect-ratio", TRUE,
 		      NULL);
-	g_object_unref (settings);
 
 	bus = gst_pipeline_get_bus (GST_PIPELINE (self->priv->playbin));
 	gst_bus_add_signal_watch (bus);
@@ -914,6 +930,13 @@ gth_media_viewer_page_real_activate (GthViewerPage *base,
 							   _("Take a screenshot"),
 							   "win.video-screenshot",
 							   NULL);
+	self->priv->fit_button =
+			gth_browser_add_header_bar_toggle_button (browser,
+								  GTH_BROWSER_HEADER_SECTION_VIEWER_ZOOM,
+								  "view-zoom-fit-symbolic",
+								  _("Fit to window"),
+								  "win.video-zoom-fit",
+								  NULL);
 
 	/* audio area */
 
@@ -934,18 +957,6 @@ gth_media_viewer_page_real_activate (GthViewerPage *base,
 			  "draw",
 			  G_CALLBACK (video_area_draw_cb),
 			  self);
-	g_signal_connect (G_OBJECT (self->priv->audio_area),
-			  "button_press_event",
-			  G_CALLBACK (video_area_button_press_cb),
-			  self);
-	g_signal_connect (G_OBJECT (self->priv->audio_area),
-			  "popup-menu",
-			  G_CALLBACK (video_area_popup_menu_cb),
-			  self);
-	g_signal_connect (G_OBJECT (self->priv->audio_area),
-			  "scroll_event",
-			  G_CALLBACK (video_area_scroll_event_cb),
-			  self);
 
 	/* mediabar */
 
@@ -996,9 +1007,23 @@ gth_media_viewer_page_real_activate (GthViewerPage *base,
 	gtk_container_add (GTK_CONTAINER (self->priv->mediabar_revealer), self->priv->mediabar);
 
 	self->priv->area_box = gtk_stack_new ();
+	gtk_style_context_add_class (gtk_widget_get_style_context (self->priv->area_box), "video-player");
 	gtk_stack_add_named (GTK_STACK (self->priv->area_box), self->priv->audio_area, "audio-area");
 	gtk_widget_show (self->priv->area_box);
 
+	g_signal_connect (G_OBJECT (self->priv->area_box),
+			  "button_press_event",
+			  G_CALLBACK (video_area_button_press_cb),
+			  self);
+	g_signal_connect (G_OBJECT (self->priv->area_box),
+			  "popup-menu",
+			  G_CALLBACK (video_area_popup_menu_cb),
+			  self);
+	g_signal_connect (G_OBJECT (self->priv->area_box),
+			  "scroll_event",
+			  G_CALLBACK (video_area_scroll_event_cb),
+			  self);
+
 	self->priv->area_overlay = gtk_overlay_new ();
 	gtk_container_add (GTK_CONTAINER (self->priv->area_overlay), self->priv->area_box);
 	gtk_overlay_add_overlay (GTK_OVERLAY (self->priv->area_overlay), self->priv->mediabar_revealer);
@@ -1013,20 +1038,8 @@ gth_media_viewer_page_real_activate (GthViewerPage *base,
 	gth_viewer_page_focus (GTH_VIEWER_PAGE (self));
 
 	create_playbin (self);
-}
 
-
-static void
-save_volume (GthMediaViewerPage *self)
-{
-	GSettings *settings;
-	double     volume;
-
-	settings = g_settings_new (GTHUMB_GSTREAMER_TOOLS_SCHEMA);
-	g_object_get (self->priv->playbin, "volume", &volume, NULL);
-	g_settings_set_int (settings, PREF_GSTREAMER_TOOLS_VOLUME, (int) (volume * 100.0));
-
-	g_object_unref (settings);
+	gth_media_viewer_page_set_fit_if_larger (self, g_settings_get_boolean (self->priv->settings, PREF_GSTREAMER_ZOOM_TO_FIT));
 }
 
 
@@ -1066,10 +1079,15 @@ gth_media_viewer_page_real_deactivate (GthViewerPage *base)
         }
 
 	if (self->priv->playbin != NULL) {
-		save_volume (self);
+		double volume;
 
-	        g_signal_handlers_disconnect_by_data (self->priv->playbin, self);
-	        g_signal_handlers_disconnect_by_data (self->priv->video_area, self);
+		g_object_get (self->priv->playbin, "volume", &volume, NULL);
+		g_settings_set_int (self->priv->settings, PREF_GSTREAMER_TOOLS_VOLUME, (int) (volume * 100.0));
+
+		g_settings_set_boolean (self->priv->settings, PREF_GSTREAMER_ZOOM_TO_FIT, self->priv->fit_if_larger);
+
+		_g_signal_handlers_disconnect_by_data (self->priv->playbin, self);
+		_g_signal_handlers_disconnect_by_data (self->priv->video_area, self);
 
 		gst_element_set_state (self->priv->playbin, GST_STATE_NULL);
 		wait_playbin_state_change_to_complete (self);
@@ -1080,7 +1098,9 @@ gth_media_viewer_page_real_deactivate (GthViewerPage *base)
 	}
 
 	gtk_widget_destroy (self->priv->screenshot_button);
+	gtk_widget_destroy (self->priv->fit_button);
 	self->priv->screenshot_button = NULL;
+	self->priv->fit_button = NULL;
 
 	gth_browser_set_viewer_widget (self->priv->browser, NULL);
 }
@@ -1251,6 +1271,7 @@ gth_media_viewer_page_real_update_sensitivity (GthViewerPage *base)
 	gtk_widget_set_sensitive (GET_WIDGET ("volume_box"), self->priv->has_audio);
 	gtk_widget_set_sensitive (GET_WIDGET ("play_button"), self->priv->has_video || self->priv->has_audio);
 	gth_window_enable_action (GTH_WINDOW (self->priv->browser), "video-screenshot", self->priv->has_video);
+	gth_window_enable_action (GTH_WINDOW (self->priv->browser), "video-zoom-fit", self->priv->has_video);
 }
 
 
@@ -1325,6 +1346,7 @@ gth_media_viewer_page_finalize (GObject *obj)
 		gth_screensaver_uninhibit (self->priv->screensaver);
 		g_object_unref (self->priv->screensaver);
 	}
+	_g_object_unref (self->priv->settings);
 
 	G_OBJECT_CLASS (gth_media_viewer_page_parent_class)->finalize (obj);
 }
@@ -1358,10 +1380,21 @@ gth_viewer_page_interface_init (GthViewerPageInterface *iface)
 }
 
 
+static void
+pref_zoom_to_fit_changed (GSettings *settings,
+			  char      *key,
+			  gpointer   user_data)
+{
+	GthMediaViewerPage *self = user_data;
+	gth_media_viewer_page_set_fit_if_larger (self, g_settings_get_boolean (self->priv->settings, PREF_GSTREAMER_ZOOM_TO_FIT));
+}
+
+
 static void
 gth_media_viewer_page_init (GthMediaViewerPage *self)
 {
 	self->priv = gth_media_viewer_page_get_instance_private (self);
+	self->priv->settings = g_settings_new (GTHUMB_GSTREAMER_TOOLS_SCHEMA);
 	self->priv->update_progress_id = 0;
 	self->priv->update_volume_id = 0;
 	self->priv->has_video = FALSE;
@@ -1377,6 +1410,14 @@ gth_media_viewer_page_init (GthMediaViewerPage *self)
 	self->priv->file_data = NULL;
 	self->priv->updated_info = NULL;
 	self->priv->loop = FALSE;
+	self->priv->fit_if_larger = TRUE;
+
+	/* settings notifications */
+
+	g_signal_connect (self->priv->settings,
+			  "changed::" PREF_GSTREAMER_ZOOM_TO_FIT,
+			  G_CALLBACK (pref_zoom_to_fit_changed),
+			  self);
 }
 
 
@@ -1418,3 +1459,58 @@ gth_media_viewer_page_get_file_data (GthMediaViewerPage *self)
 {
 	return self->priv->file_data;
 }
+
+
+void
+gth_media_viewer_page_toggle_play (GthMediaViewerPage *self)
+{
+	if (self->priv->playbin == NULL)
+		return;
+
+	if (! self->priv->playing) {
+		if (! self->priv->paused) {
+			gst_element_set_state (self->priv->playbin, GST_STATE_PAUSED);
+			gst_element_seek (self->priv->playbin,
+					  self->priv->rate,
+					  GST_FORMAT_TIME,
+					  GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_ACCURATE,
+					  GST_SEEK_TYPE_SET,
+					  0.0,
+					  GST_SEEK_TYPE_NONE,
+					  0.0);
+		}
+		else {
+			gint64 current_value;
+
+			current_value = (gint64) (gtk_adjustment_get_value (GTK_ADJUSTMENT (GET_WIDGET ("position_adjustment"))) / 100.0 * self->priv->duration);
+			gst_element_seek (self->priv->playbin,
+					  self->priv->rate,
+					  GST_FORMAT_TIME,
+					  GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_ACCURATE,
+					  GST_SEEK_TYPE_SET,
+					  current_value,
+					  GST_SEEK_TYPE_NONE,
+					  0.0);
+		}
+		gst_element_set_state (self->priv->playbin, GST_STATE_PLAYING);
+	}
+	else
+		gst_element_set_state (self->priv->playbin, GST_STATE_PAUSED);
+}
+
+
+void
+gth_media_viewer_page_set_fit_if_larger (GthMediaViewerPage *self,
+					 gboolean            fit_if_larger)
+{
+	GtkAlign alignment;
+
+	self->priv->fit_if_larger = fit_if_larger;
+	if (self->priv->video_area != NULL) {
+		alignment = self->priv->fit_if_larger ? GTK_ALIGN_FILL : GTK_ALIGN_CENTER;
+		gtk_widget_set_valign (self->priv->video_area, alignment);
+		gtk_widget_set_halign (self->priv->video_area, alignment);
+
+		gth_window_change_action_state (GTH_WINDOW (self->priv->browser), "video-zoom-fit", self->priv->fit_if_larger);
+	}
+}
